# $Id: absclient.py,v 1.7 2002/02/15 10:33:57 marijn Exp $

# An air traffic simulation game. 
# Copyright (C) 2000,1,2  Marijn Vriens <marijn@sanity.dhs.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import threading
import string
import CORBA
import time

import setup

class AbsClient(threading.Thread):
    """This class is designed to be inherited by other classes
    you cant use it on it's own. This class provides UI, network threading
    intergration"""

    def __init__(self, fifoWrite):
        #print "%s ctor" % (self.__class__)        
        threading.Thread.__init__(self)

        self._fifo = fifoWrite
        
        # generate a list with *_impl methodes
        self._methodDict = {}
        for e in list(dir(self.__class__)):
            p = string.find(e, '_impl')
            if(p != -1):
                self._methodDict[e[:p]] = getattr(self, e)
                #print "%s --> %s" % (e, e[:p])
            self._event = threading.Event()
        self._eventList = [] # Queue with things for the thread to do.
        self._returnList= [] # Queue with return data for the caller.
        self._listsLock = threading.Lock()

        self._eventDelayer = DelayEvents()

        self._exceptionHandlerNameDict = {}
        self._exceptionHandlerClassDict = {}

        self._genExceptionNameClassLink()

        self._exitRequested = 0

        self._orb = CORBA.ORB_init()
        self.setDaemon(1) # quit even if thread still running.
        #print "Starting thread"
        self.start()

    def run(self):
        #print "Done"
        # open a fifo to signal the gtk-thread.
        #print "absthread pre-fifo open"
        self._notifyUi()
        #print "absthread post-fifo open"
        while(1):
            self._event.wait() # wait for an event to happen.
            #print "AbsClient::run: passed through _event.wait()"
            if(self._exitRequested):
                self._doExit()
                return
            events  = self._eventList
            returns = self._returnList
            lock = self._listsLock
            while(events):
                #print "AbsClient::run: waiting to aquiring lock" 
                lock.acquire()
                #print "AbsClient::run: aquired lock"
                e = events.pop()
                lock.release()
                #print "AbsClient::run: released lock, doing event"
                e.doEvent()
                #print "AbsClient::run: event done, aquireing lock"
                lock.acquire()
                #print "AbsClient::run: aquired lock"
                returns.insert(0, e)
                lock.release()
                #print "AbsClient::run: released lock"
                self._notifyUi()
            self._event.clear() # block event.wait().
    def __getattr__(self, attrname):
        # Called by UI to get work schedueled in the thread.
        try:
            realMethod = self._methodDict[attrname] # generated in __init__()
        except(KeyError):
            raise AttributeError, attrname
        #print "AbsClient::__getattr__: %s : %s" % (attrname, realMethod)
        proxy = _EventObject(self._eventList, self._listsLock, self._event,
                             self._checkException, realMethod)
        return proxy

    def executeWithDelay(self, delay, attrname, callback, *opts):
        # Called by UI to get work schedueled after delay seconds in the
        # thread.
        proxy = self.__getattr__(attrname)
        self._eventDelayer.addEvent(delay, proxy, callback, opts)

    def doCallBacks(self):
        # Called by the UI to execute the callbacks.
        while(self._returnList):
            self._listsLock.acquire()
            e = self._returnList.pop()
            self._listsLock.release()
            e.doCallback()

    def addExceptionHandler(self, exceptionName, handler):
        # Called by the Ui to add a handler.
        self._exceptionHandlerNameDict[exceptionName] = handler

    def requestExit(self):
        # called by the Ui to exit the thread. 
        self._exitRequested = 1
        self._event.set()

    def _doExit(self):
            #print "Thread %s: %s exited as requested" % (repr(id(self)),
            #                                              self)
            self._notifyUi()
            #os.close(self._fifo)

    def _notifyUi(self):
        str = "%s " % repr(id(self))
        #print "AbsClient::_notifyUi: Write %s to fifo" % str
        os.write(self._fifo, str)
        #print "AbsClient::_notifyUi: Done"

    def _checkException(self, eClass):
        try:
            #print self._exceptionHandlerClassDict
            methodName = self._exceptionHandlerClassDict[eClass]
        except KeyError:
            return self._unknownExceptionCall
        try:
            #print "_checkException", self._exceptionHandlerNameDict
            return self._exceptionHandlerNameDict[methodName]
        except KeyError:
            return self._unhandledExceptionCall

    def _genExceptionNameClassLink(self):
        self._exceptionHandlerClassDict[CORBA.COMM_FAILURE] = "CORBA_Failure"
    def _unknownExceptionCall(self, e):
        raise "AbstractMethodCalled::_unknownExceptionCall", e.__class__
    def _unhandledExceptionCall(self, e):
        raise "AbstractMethodCalled::_unhandledExceptionCall", e.__class__



class _EventObject:
    def __init__(self, list, lock, event, exceptionHandlerList, method):
        self.list = list
        self.lock = lock
        self.event= event
        self.method = method
        self.rvalue = None
        self.exceptionList = exceptionHandlerList
        self.methDone = 0
    def __call__(self, callback, *opts):
        #print "_EventObject::__call__: %s : %s" %(callback, opts)
        self.callback = callback
        self.methargs = opts
        self.lock.acquire()
        self.list.insert(0, self)
        self.lock.release()
        self.event.set()
    def doEvent(self): # do the event in the running thread.
        #print "doing EVENT: %s : %s" % (self.method, self.methargs)
        try:
            self.rvalue = apply(self.method, self.methargs)
        except (CORBA.SystemException, CORBA.UserException), e:
            # okay, so we got an exception, so, lets change the
            # the UI callback to handle that. 
            #print "_EventObject::doEvent: Got exception"
            #print "Class = ", e.__class__
            self.callback = self.exceptionList(e.__class__)
            self.rvalue = e
    def doCallback(self): # do the callback in the calling thread.
        #print "doing CALLBACK: %s : %s" % (self.callback, self.rvalue)
        if self.callback == None: return
        if not self.rvalue:
            r = apply(self.callback)
        elif type(self.rvalue) == type(()) or type(self.rvalue) == type([]):
            r = apply(self.callback, self.rvalue)
        else:
            r = apply(self.callback, (self.rvalue,))

        if r:
            print "Are you sure wanted to return? We're not listening!"
            raise "NoReturnExpected"
    def setTimeOfExecution(self, delay):
        self._timeOfExecution = time.time() + delay
    def timeOfExecutionPassed(self, timeNow):
        if timeNow > self._timeOfExecution:
            return self
        return None



def DelayEvents():
    if _DelayEvents.instance == None:
        _DelayEvents.instance = _DelayEvents()
    return _DelayEvents.instance

class _DelayEvents(threading.Thread):
    instance = None
    def __init__(self):
        threading.Thread.__init__(self)
        self._waitingList = []
        self._iterationSpeed = 0.25 
        self.start()
    def run(self):
        while(1):
            l = self._getEventsPending()
            if(l):
                self._callEvents(l)
            time.sleep(self._iterationSpeed)
    def addEvent(self, delay, eventObj, callback, opts):
        data = (eventObj, callback,) + opts
        self._waitingList.append(data)
        eventObj.setTimeOfExecution(delay)
    def _getEventsPending(self):
        timeNow = time.time()
        l = []
        for i in self._waitingList:
            if i[0].timeOfExecutionPassed(timeNow):
                l.append(i)
                del self._waitingList[l.index(i)]
        return l
    def _callEvents(self, list):
        for i in list:
            apply(i[0], i[1:])

# $Log: absclient.py,v $
# Revision 1.7  2002/02/15 10:33:57  marijn
# Added network exception handling framework
#
# Revision 1.6  2002/02/02 17:23:57  marijn
# Updated copyright information
#
# Revision 1.5  2002/01/24 14:10:18  marijn
# Silenced overtly verbose message
#
# Revision 1.4  2002/01/16 18:09:54  marijn
# Changed from named-pipes to normal ones.
# Added AbsClient.executeWithDelay() that executes the event later.
# Fixed some inter-threading bugs. the Exception de-syncing is still here.
#
# Revision 1.3  2002/01/14 15:50:26  marijn
# Added network to Ui support via Fifo. General Cleaning up
#
# Revision 1.2  2001/12/26 02:33:03  marijn
# removed some verbosity
#
# Revision 1.1  2001/12/11 03:53:23  marijn
# Split code off, moved common client code to absclient.py
#
